from typing import Optional, Union

import pytest

from exotx.data.static import Calendar, CalendarRegion, CalendarMarket


# region from_json
@pytest.mark.parametrize('region, market', [
    ('UnitedStates', 'NYSE'),
    (CalendarRegion.UnitedStates, CalendarMarket.NYSE),
    (CalendarRegion.UnitedStates, None)
])
def test_calendar_from_json(region: Union[CalendarRegion, str, None], market: Union[CalendarMarket, str, None]):
    # Arrange
    if isinstance(region, str):
        my_region = region
    elif isinstance(region, CalendarRegion):
        my_region = region.name
    else:
        my_region = None

    if isinstance(market, str):
        my_market = market
    elif isinstance(market, CalendarMarket):
        my_market = market.name
    else:
        my_market = None

    my_json = {
        'region': my_region,
        'market': my_market
    }

    # Act
    my_calendar = Calendar.from_json(my_json)

    # Assert
    assert isinstance(my_calendar.region, CalendarRegion)
    assert isinstance(my_calendar.market, CalendarMarket)
    if region:
        if isinstance(region, str):
            assert my_calendar.region.name == region
        else:
            assert my_calendar.region == region
        if market:
            if isinstance(market, str):
                assert my_calendar.market.name == market
            else:
                assert my_calendar.market == market
        else:
            assert my_calendar.market == CalendarMarket.Settlement
    else:
        assert my_calendar.region is None
        assert my_calendar.market is None


@pytest.mark.parametrize('region', [
    'UnitedStates'
])
def test_calendar_from_json_default_market(region: Union[CalendarRegion, str, None]):
    # Arrange
    my_json = {
        'region': region
    }

    # Act
    my_calendar = Calendar.from_json(my_json)

    # Assert
    assert isinstance(my_calendar.region, CalendarRegion)
    if isinstance(region, str):
        assert my_calendar.region.name == region
    elif isinstance(region, CalendarRegion):
        assert my_calendar.region == region
    assert my_calendar.market == CalendarMarket.Settlement


@pytest.mark.parametrize('market', [
    'NYSE'
])
def test_calendar_from_json_missing_region_for_market(market: Optional[str]):
    # Arrange
    my_json = {
        'market': market
    }

    # Act
    with pytest.raises(Exception) as e:
        _ = Calendar.from_json(my_json)

    # Assert
    assert str(e.value) == f'Missing region for market \'{market}\''


@pytest.mark.parametrize('region, market', [
    ('UnitedState', 'NYSE')  # Invalid region
])
def test_calendar_from_json_invalid_region(region: Optional[str], market: Optional[str]):
    # Arrange
    my_json = {
        'region': region,
        'market': market,
    }

    # Act
    with pytest.raises(Exception) as e:
        _ = Calendar.from_json(my_json)

    # Assert
    assert str(e.value) == f'Invalid region \'{region}\'. Available regions: Argentina, Australia, BespokeCalendar, ' \
                           f'Brazil, Canada, China, CzechRepublic, Denmark, Finland, France, Germany, HongKong, ' \
                           f'Hungary, Iceland, India, Indonesia, Israel, Italy, Japan, Mexico, NewZealand, Norway, ' \
                           f'NullCalendar, Poland, Romania, Russia, SaudiArabia, Singapore, Slovakia, SouthAfrica, ' \
                           f'SouthKorea, Sweden, Switzerland, Taiwan, TARGET, Thailand, Turkey, Ukraine, UnitedKingdom, ' \
                           f'UnitedStates, WeekendsOnly'


@pytest.mark.parametrize('region, market', [
    ('UnitedStates', 'whatever')  # Invalid market for that region
])
def test_calendar_from_json_invalid_market_for_region(region: Optional[str], market: Optional[str]):
    # Arrange
    my_json = {
        'region': region,
        'market': market,
    }

    # Act
    with pytest.raises(Exception) as e:
        _ = Calendar.from_json(my_json)

    # Assert
    assert str(e.value) == f'Invalid market \'{market}\' for region \'{region}\'. Available markets: Settlement, ' \
                           f'FederalReserve, GovernmentBond, LiborImpact, NERC, NYSE'


# endregion


# region to_json
@pytest.mark.parametrize('region, market', [
    ('UnitedStates', 'NYSE'),
    (CalendarRegion.UnitedStates, CalendarMarket.NYSE),
    (CalendarRegion.UnitedStates, None)
])
def test_calendar_to_json(region: Union[CalendarRegion, str, None], market: Union[CalendarMarket, str, None]):
    # Arrange
    my_calendar = Calendar(region=region, market=market)

    # Act
    my_json = my_calendar.to_json()

    # Assert
    if isinstance(region, str):
        assert my_json['region'] == region
    elif isinstance(region, CalendarRegion):
        assert my_json['region'] == region.name
    if market is None:
        assert my_json['market'] == 'Settlement'
    else:
        if isinstance(market, str):
            assert my_json['market'] == market
        elif isinstance(market, CalendarMarket):
            assert my_json['market'] == market.name

# endregion



from enum import Enum
from typing import Union

from marshmallow import fields, ValidationError


class AverageConvention(Enum):
    PRICE = 'price'
    STRIKE = 'strike'


def convert_average_convention(average_convention: Union[str, AverageConvention]) -> AverageConvention:
    if isinstance(average_convention, str):
        average_convention = average_convention.upper()
        if average_convention not in AverageConvention.__members__:
            raise ValueError(
                f"Invalid average convention \"{average_convention}\", expected one of {list(AverageConvention.__members__.keys())}")
        return AverageConvention[average_convention]
    elif isinstance(average_convention, AverageConvention):
        return average_convention
    else:
        raise Exception(f"Invalid average convention type \"{type(average_convention)}\"")


class AverageConventionField(fields.Field):
    def _serialize(self, value: AverageConvention, attr, obj, **kwargs) -> str:
        return value.name

    def _deserialize(self, value: str, attr, data, **kwargs) -> AverageConvention:
        try:
            return AverageConvention[value]
        except KeyError as error:
            raise ValidationError(
                f"Invalid average convention \"{value}\", expected one of {list(AverageConvention.__members__.keys())}") from error



from exotx.data.static.calendar import Calendar
from exotx.data.static.calendars import calendars_to_ql_calendars, CalendarRegion, CalendarMarket, region_to_markets, \
    available_regions
from exotx.data.static.conventions import compoundings, frequencies, business_day_conventions_to_ql, \
    BusinessDayConvention
from exotx.data.static.daycounters import day_counters_to_ql, DayCounter

__all__ = [
    'calendars_to_ql_calendars',
    'day_counters_to_ql',
    'compoundings',
    'frequencies',
    'business_day_conventions_to_ql',
    'BusinessDayConvention',
    'Calendar',
    'CalendarRegion',
    'CalendarMarket',
    'region_to_markets',
    'available_regions',
    'DayCounter'
]



import json
from datetime import datetime
from typing import List, Union

import QuantLib as ql
from marshmallow import Schema, fields, post_load


class MarketData:

    def __init__(self,
                 underlying_spots: List[float],
                 risk_free_rate: float,
                 dividend_rate: float,
                 reference_date: Union[datetime, str, None] = None,
                 expiration_dates: List[Union[datetime, str]] = None,
                 strikes: List[float] = None,
                 data: List[List[float]] = None,
                 underlying_black_scholes_volatilities: List[float] = None,
                 correlation_matrix: List[List[float]] = None) -> None:
        # set the reference date
        self._set_reference_date(reference_date)

        # set underlying spot values
        self._set_underlying_spots(underlying_spots)

        # set market rate curves
        self._set_rate_curves(dividend_rate, risk_free_rate)

        # set the volatility surface
        self._set_volatility_surface(underlying_black_scholes_volatilities, data, expiration_dates, strikes)

        # set the correlation matrix
        self._set_correlation_matrix(correlation_matrix)

    # region setters

    # TODO: Have a proper rate curves stripper service
    def _set_rate_curves(self, dividend_rate, risk_free_rate) -> None:
        self.risk_free_rate = risk_free_rate
        self.dividend_rate = dividend_rate

    # TODO: Allow for multiple volatility surfaces for each underlying
    def _set_volatility_surface(self, underlying_black_scholes_volatilities: List[float], data, expiration_dates,
                                strikes) -> None:
        self.expiration_dates: List[datetime] = []
        if not expiration_dates:
            self.expiration_dates = None
        else:
            for expiration_date in expiration_dates:
                if isinstance(expiration_date, str):
                    self.expiration_dates.append(datetime.strptime(expiration_date, '%Y-%m-%d'))
                elif isinstance(expiration_date, datetime):
                    self.expiration_dates.append(expiration_date)
        self.strikes = strikes
        self.data = data

        if underlying_black_scholes_volatilities:
            for vol in underlying_black_scholes_volatilities:
                assert vol >= 0, f"Invalid volatility: {vol}"
        self.underlying_black_scholes_volatilities = underlying_black_scholes_volatilities  # set to None if it does not exist, OK

    # TODO: Allow for multiple underlying spots to be defined, may need a proper container class or a dict
    def _set_underlying_spots(self, underlying_spots: List[float]) -> None:
        for underlying_spot in underlying_spots:
            assert underlying_spot > 0, f"Invalid underlying spot {underlying_spot}"
        self.underlying_spots = underlying_spots

    def _set_reference_date(self, reference_date: Union[datetime, str, None]) -> None:
        if isinstance(reference_date, str):
            reference_date = datetime.strptime(reference_date, '%Y-%m-%d')
        elif isinstance(reference_date, datetime):
            pass
        else:
            # defaults to today's date
            reference_date = datetime.today()
        self.reference_date: datetime = reference_date

    def _set_correlation_matrix(self, correlation_matrix: Union[List[List[float]], None]) -> None:
        if correlation_matrix:
            for rows in correlation_matrix:
                for rho in rows:
                    assert 1 >= rho >= -1, "Invalid correlation matrix"
        self.correlation_matrix = correlation_matrix

    # endregion

    # region getters
    def get_ql_reference_date(self) -> ql.Date:
        return ql.Date().from_date(self.reference_date)

    def get_correlation_matrix(self) -> ql.Matrix:
        matrix = ql.Matrix(len(self.correlation_matrix), len(self.correlation_matrix))
        for i in range(len(self.correlation_matrix)):
            matrix[i][i] = 1.0
            for j in range(i + 1, len(self.correlation_matrix)):
                matrix[i][j] = self.correlation_matrix[i][j]
                matrix[j][i] = self.correlation_matrix[i][j]
        return matrix

    # TODO: Get these from a proper rate curve stripper service
    def get_yield_curve(self, day_counter) -> ql.YieldTermStructureHandle:
        flat_forward = ql.FlatForward(self.get_ql_reference_date(), self.risk_free_rate, day_counter)
        return ql.YieldTermStructureHandle(flat_forward)

    def get_dividend_curve(self, day_counter) -> ql.YieldTermStructureHandle:
        flat_forward = ql.FlatForward(self.get_ql_reference_date(), self.dividend_rate, day_counter)
        return ql.YieldTermStructureHandle(flat_forward)

    # endregion

    # region serialization/deserialization
    @classmethod
    def from_json(cls, data: dict):
        schema = MarketDataSchema()
        return schema.load(data)

    def to_json(self, format_type: str = "dict"):
        schema = MarketDataSchema()
        my_json = schema.dump(self)
        if format_type == "dict":
            return my_json
        elif format_type == "str":
            return json.dumps(my_json)
        else:
            raise NotImplementedError(f"Invalid format type {format_type} when dumping")
    # endregion


# region Schema

class MarketDataSchema(Schema):
    underlying_spots = fields.List(fields.Float(required=True))
    risk_free_rate = fields.Float(required=True)
    dividend_rate = fields.Float(required=True)
    reference_date = fields.DateTime(format='%Y-%m-%d', allow_none=True)
    expiration_dates = fields.List(fields.DateTime(format='%Y-%m-%d'), allow_none=True)
    strikes = fields.List(fields.Float(), allow_none=True)
    data = fields.List(fields.List(fields.Float), allow_none=True)
    underlying_black_scholes_volatilities = fields.List(fields.Float(allow_none=True))
    correlation_matrix = fields.List(fields.List(fields.Float()))

    @post_load
    def make_market_data(self, data, **kwargs) -> MarketData:
        return MarketData(**data)

# endregion



from abc import ABC, abstractmethod
from typing import Union


class Instrument(ABC):
    @abstractmethod
    def price(self, *args, **kwargs) -> Union[float, dict]:
        pass


def price(instrument: Instrument, *args, **kwargs) -> Union[float, dict]:
    return instrument.price(*args, **kwargs)



from datetime import datetime
from typing import Union

import QuantLib as ql
from marshmallow import Schema, fields, post_load

from exotx.enums.enums import PricingModel, NumericalMethod
from exotx.helpers.dates import convert_maturity_to_ql_date
from exotx.instruments.instrument import Instrument
from exotx.instruments.option_type import convert_option_type_to_ql, OptionType, OptionTypeField
from exotx.models.blackscholesmodel import BlackScholesModel
from exotx.models.hestonmodel import HestonModel
from exotx.utils.pricing_configuration import PricingConfiguration


class VanillaOption(Instrument):
    available_engines = {
        PricingModel.BLACK_SCHOLES: ql.AnalyticEuropeanEngine,
        PricingModel.HESTON: ql.AnalyticHestonEngine
    }

    def __init__(self, strike: float, maturity: Union[str, datetime], option_type: Union[str, OptionType]):
        assert strike >= 0, "Invalid strike: cannot be negative"
        self.strike = strike
        self.maturity = convert_maturity_to_ql_date(maturity)
        self.option_type = convert_option_type_to_ql(option_type)

    def price(self, market_data, static_data, pricing_config: PricingConfiguration, seed: int = 1) -> dict:
        reference_date: ql.Date = market_data.get_ql_reference_date()
        ql.Settings.instance().evaluationDate = reference_date

        # create the product
        ql_payoff = ql.PlainVanillaPayoff(self.option_type, self.strike)
        ql_exercise = ql.EuropeanExercise(self.maturity)
        ql_option = ql.VanillaOption(ql_payoff, ql_exercise)

        # set the pricing engine
        ql_engine = self._get_ql_pricing_engine(market_data, static_data, pricing_config, seed)
        ql_option.setPricingEngine(ql_engine)

        # price
        price = ql_option.NPV()
        if pricing_config.compute_greeks and pricing_config.model != PricingModel.HESTON:
            delta = ql_option.delta()
            gamma = ql_option.gamma()
            theta = ql_option.theta()
            return {'price': price, 'delta': delta, 'gamma': gamma, 'theta': theta}
        else:
            return {'price': price}

    @staticmethod
    def _get_ql_pricing_engine(market_data, static_data, pricing_config: PricingConfiguration, seed: int):
        if pricing_config.model == PricingModel.BLACK_SCHOLES and \
                pricing_config.numerical_method == NumericalMethod.ANALYTIC:
            bs_model = BlackScholesModel(market_data, static_data)
            process = bs_model.setup()
            ql_engine = ql.AnalyticEuropeanEngine(process)
        elif pricing_config.model == PricingModel.HESTON and \
                pricing_config.numerical_method == NumericalMethod.ANALYTIC:
            heston_model = HestonModel(market_data, static_data)
            _, model = heston_model.calibrate(seed)
            ql_engine = ql.AnalyticHestonEngine(model)
        else:
            raise ValueError(f"Invalid pricing model {pricing_config.model} with numerical method "
                             f"{pricing_config.numerical_method}")

        return ql_engine

    # region serialization/deserialization
    def to_json(self):
        schema = VanillaOptionSchema()
        return schema.dump(self)

    @classmethod
    def from_json(cls, json_data):
        schema = VanillaOptionSchema()
        data = schema.load(json_data)
        return cls(**data)
    # endregion


# region Schema


class VanillaOptionSchema(Schema):
    strike = fields.Float()
    maturity = fields.Date(format="%Y-%m-%d")
    option_type = OptionTypeField(allow_none=False)

    @post_load
    def make_vanilla_option(self, data, **kwargs) -> VanillaOption:
        return VanillaOption(**data)

# endregion



from enum import Enum
from typing import Union

import QuantLib as ql
from marshmallow import fields, ValidationError


class AverageType(Enum):
    ARITHMETIC = 'arithmetic'
    GEOMETRIC = 'geometric'


def convert_average_type_to_ql(average_type: Union[str, AverageType]):
    if isinstance(average_type, str):
        average_type = average_type.upper()
        if average_type not in AverageType.__members__:
            raise ValueError(
                f"Invalid average type \"{average_type}\", expected one of {list(AverageType.__members__.keys())}")
        return convert_average_type_to_ql(AverageType[average_type])
    elif isinstance(average_type, AverageType):
        if average_type == AverageType.ARITHMETIC:
            return ql.Average().Arithmetic
        elif average_type == AverageType.GEOMETRIC:
            return ql.Average().Geometric
        else:
            raise ValueError(f"Invalid average type \"{average_type}\"")
    else:
        raise TypeError(f"Invalid input type {type(average_type)} for average type")


class AverageTypeField(fields.Field):
    def _serialize(self, value: AverageType, attr, obj, **kwargs) -> str:
        return value.name

    def _deserialize(self, value: str, attr, data, **kwargs) -> AverageType:
        try:
            return AverageType[value]
        except KeyError as error:
            raise ValidationError(
                f"Invalid average type \"{value}\", expected one of {list(AverageType.__members__.keys())}") from error



from enum import Enum
from typing import Union

from marshmallow import fields, ValidationError


class BasketType(Enum):
    MINBASKET = 'minbasket'
    MAXBASKET = 'maxbasket'
    SPREADBASKET = 'spreadbasket'
    AVERAGEBASKET = 'averagebasket'


def convert_basket_type(basket_type: Union[str, BasketType]) -> BasketType:
    if isinstance(basket_type, str):
        basket_type = basket_type.upper()
        if basket_type not in BasketType.__members__:
            raise ValueError(
                f"Invalid basket type \"{basket_type}\", expected one of {list(BasketType.__members__.keys())}")
        return BasketType[basket_type]
    elif isinstance(basket_type, BasketType):
        return basket_type
    else:
        raise Exception(f"Invalid basket type \"{type(basket_type)}\"")


class BasketTypeField(fields.Field):
    def _serialize(self, value: BasketType, attr, obj, **kwargs) -> str:
        return value.name

    def _deserialize(self, value: str, attr, data, **kwargs) -> BasketType:
        try:
            return BasketType[value]
        except KeyError as error:
            raise ValidationError(
                f"Invalid basket type \"{value}\", expected one of {list(BasketType.__members__.keys())}") from error



from enum import Enum
from typing import Union

import QuantLib as ql
from marshmallow import fields, ValidationError


class OptionType(Enum):
    CALL = 'call'
    PUT = 'put'

    @staticmethod
    def values():
        return [e.value for e in OptionType]


def convert_option_type_to_ql(option_type: Union[str, OptionType]) -> ql.Option:
    if isinstance(option_type, str):
        option_type = option_type.upper()
        if option_type not in OptionType.__members__:
            raise ValueError(f"Invalid option type \"{option_type}\", expected one of "
                             f"{list(OptionType.__members__.keys())}")
        return convert_option_type_to_ql(OptionType[option_type])
    elif isinstance(option_type, OptionType):
        if option_type == OptionType.CALL:
            return ql.Option.Call
        elif option_type == OptionType.PUT:
            return ql.Option.Put
        else:
            raise ValueError(f"Invalid option type \"{option_type}\"")
    else:
        raise TypeError(f"Invalid input type {type(option_type)} for option type")


class OptionTypeField(fields.Field):
    def _serialize(self, value: OptionType, attr, obj, **kwargs) -> str:
        return value.name

    def _deserialize(self, value: str, attr, data, **kwargs) -> OptionType:
        try:
            return OptionType[value]
        except KeyError as error:
            raise ValidationError(f"Invalid option type \"{value}\", expected one of "
                                  f"{list(OptionType.__members__.keys())}") from error



from datetime import timedelta

import QuantLib as ql
import pytest

from exotx import price
from exotx.data.marketdata import MarketData
from exotx.data.staticdata import StaticData
from exotx.enums.enums import PricingModel, NumericalMethod
from exotx.instruments.asian_option import AsianOption, AverageCalculation, AverageConvention
from exotx.instruments.average_type import AverageType
from exotx.instruments.option_type import OptionType
from exotx.utils.pricing_configuration import PricingConfiguration


# Arrange
@pytest.fixture()
def my_static_data() -> StaticData:
    my_json = {
        'day_counter': 'Actual360',
        'business_day_convention': 'ModifiedFollowing'
    }
    return StaticData.from_json(my_json)


@pytest.fixture()
def my_market_data() -> MarketData:
    my_json = {
        'reference_date': '2015-11-06',
        'underlying_spots': [80],
        'risk_free_rate': 0.05,
        'dividend_rate': -0.03,
        'underlying_black_scholes_volatilities': [0.20]
    }

    return MarketData.from_json(my_json)


@pytest.fixture
def my_asian_option() -> AsianOption:
    strike = 85
    maturity = '2016-02-04'  # reference_date + 90 days
    exercise = 'european'
    option_type = OptionType.PUT
    # option_type = 'call'
    average_type = AverageType.GEOMETRIC
    # average_type = 'geometric'
    average_calculation = AverageCalculation.CONTINUOUS
    average_convention = AverageConvention.PRICE

    return AsianOption(strike, maturity, option_type, average_type, average_calculation, average_convention)


@pytest.fixture
def my_pricing_config() -> PricingConfiguration:
    model = PricingModel.BLACK_SCHOLES
    numerical_method = NumericalMethod.ANALYTIC

    return PricingConfiguration(model, numerical_method)


# Reproduces https://github.com/lballabio/QuantLib/blob/master/test-suite/asianoptions.cpp#L95
@pytest.mark.parametrize(
    'average_calculation, compute_greeks, expected_price, expected_delta, expected_gamma, expected_theta',
    [
        (AverageCalculation.CONTINUOUS, True, 4.6922, -0.80308771, 0.05935118, 0.05801764)
    ])
def test_price(my_asian_option: AsianOption,
               my_market_data: MarketData,
               my_static_data: StaticData,
               my_pricing_config: PricingConfiguration,
               average_calculation: AverageCalculation,
               compute_greeks: bool,
               expected_price: float,
               expected_delta: float,
               expected_gamma: float,
               expected_theta: float) -> None:
    # Arrange
    my_asian_option.average_calculation = average_calculation
    my_pricing_config.compute_greeks = compute_greeks

    # Act
    result = price(my_asian_option, my_market_data, my_static_data, my_pricing_config)

    # Assert
    assert result['price'] == pytest.approx(expected_price, abs=1e-4)
    if my_pricing_config.compute_greeks:
        assert result['delta'] == pytest.approx(expected_delta, abs=1e-8)
        assert result['gamma'] == pytest.approx(expected_gamma, abs=1e-8)
        assert result['theta'] == pytest.approx(expected_theta, abs=1e-8)


# Reproduces https://github.com/lballabio/QuantLib/blob/master/test-suite/asianoptions.cpp#L95
@pytest.mark.parametrize(
    'average_calculation, expected_price',
    [
        (AverageCalculation.DISCRETE, 4.6922)
    ])
def test_price_approximate_continuous_with_discrete_version(my_asian_option: AsianOption,
                                                            my_market_data: MarketData,
                                                            my_static_data: StaticData,
                                                            my_pricing_config: PricingConfiguration,
                                                            average_calculation: AverageCalculation,
                                                            expected_price: float) -> None:
    # Arrange
    my_asian_option.average_calculation = average_calculation
    my_asian_option.future_fixing_dates = [ql.Date().from_date(my_market_data.reference_date + timedelta(days=i))
                                           for i in range(1, 90)]

    # Act
    result = price(my_asian_option, my_market_data, my_static_data, my_pricing_config)

    # Assert
    assert result['price'] == pytest.approx(expected_price, abs=3.0e-3)


# Reproduces https://github.com/lballabio/QuantLib/blob/master/test-suite/asianoptions.cpp#L384
@pytest.mark.parametrize(
    'average_calculation, average_convention, strike, expected_price',
    [
        (AverageCalculation.DISCRETE, AverageConvention.STRIKE, 100, 4.97109)
    ])
def test_price_analytic_discrete_geometric_average_strike(my_asian_option: AsianOption,
                                                          my_market_data: MarketData,
                                                          my_static_data: StaticData,
                                                          my_pricing_config: PricingConfiguration,
                                                          average_calculation: AverageCalculation,
                                                          average_convention: AverageConvention,
                                                          strike: float,
                                                          expected_price: float) -> None:
    # Arrange
    my_asian_option.option_type = ql.Option.Call
    my_asian_option.maturity = ql.Date(31, 10, 2016)
    my_asian_option.strike = strike
    my_asian_option.average_calculation = average_calculation
    my_asian_option.future_fixing_dates = [ql.Date().from_date(my_market_data.reference_date + timedelta(days=36 * i))
                                           for i in range(1, 11)]
    my_asian_option.average_convention = average_convention
    my_market_data.underlying_spots = [100]
    my_market_data.risk_free_rate = 0.06
    my_market_data.dividend_rate = 0.03
    my_market_data.underlying_black_scholes_volatilities = [0.20]

    # Act
    result = price(my_asian_option, my_market_data, my_static_data, my_pricing_config)

    # Assert
    assert result['price'] == pytest.approx(expected_price, abs=1e-5)



import enum

import QuantLib as ql


# based on https://quantlib-python-docs.readthedocs.io/en/latest/dates.html#daycounter


class DayCounter(enum.Enum):
    SimpleDayCounter = 0
    Thirty360 = 1
    Actual360 = 2
    Actual365Fixed = 3
    Actual365FixedCanadian = 4
    Actual365FixedNoLeap = 5
    ActualActual = 6
    Business252 = 7


day_counters_to_ql = {
    DayCounter.SimpleDayCounter: ql.SimpleDayCounter(),
    DayCounter.Thirty360: ql.Thirty360(),
    DayCounter.Actual360: ql.Actual360(),
    DayCounter.Actual365Fixed: ql.Actual365Fixed(),
    DayCounter.Actual365FixedCanadian: ql.Actual365Fixed(ql.Actual365Fixed.Canadian),
    DayCounter.Actual365FixedNoLeap: ql.Actual365Fixed(ql.Actual365Fixed.NoLeap),
    DayCounter.ActualActual: ql.ActualActual(),
    DayCounter.Business252: ql.Business252()
}



from enum import Enum
from typing import Union

from marshmallow import fields, ValidationError


class AverageCalculation(Enum):
    CONTINUOUS = 'continuous'
    DISCRETE = 'discrete'


def convert_average_calculation(average_calculation: Union[str, AverageCalculation]) -> AverageCalculation:
    if isinstance(average_calculation, str):
        average_calculation = average_calculation.upper()
        if average_calculation not in AverageCalculation.__members__:
            raise ValueError(
                f"Invalid average calculation \"{average_calculation}\", expected one of {list(AverageCalculation.__members__.keys())}")
        return AverageCalculation[average_calculation]
    elif isinstance(average_calculation, AverageCalculation):
        return average_calculation
    else:
        raise Exception(f"Invalid average calculation type \"{type(average_calculation)}\"")


class AverageCalculationField(fields.Field):
    def _serialize(self, value: AverageCalculation, attr, obj, **kwargs) -> str:
        return value.name

    def _deserialize(self, value: str, attr, data, **kwargs) -> AverageCalculation:
        try:
            return AverageCalculation[value]
        except KeyError as error:
            raise ValidationError(
                f"Invalid average calculation \"{value}\", expected one of {list(AverageCalculation.__members__.keys())}") from error



import pytest

from exotx import price
from exotx.data.marketdata import MarketData
from exotx.data.staticdata import StaticData
from exotx.instruments.barrier_option import BarrierOption, BarrierType


# replicates the tests in https://github.com/lballabio/QuantLib/blob/master/test-suite/barrieroption.cpp


# Arrange
@pytest.fixture
def my_barrier_option() -> BarrierOption:
    barrier_type = 'upandin'
    barrier = 105
    strike = 90
    maturity = '2016-05-04'  # time to maturity = 0.5
    exercise = 'european'
    option_type = 'call'
    rebate = 3.0

    return BarrierOption(barrier_type, barrier, strike, maturity, exercise, option_type, rebate)


# replicates https://github.com/lballabio/QuantLib/blob/master/test-suite/barrieroption.cpp#L280-L282
@pytest.mark.parametrize('barrier_type, strike, model, expected', [
    (BarrierType.UPANDIN, 90, 'analytic', 14.111173119603055),
    (BarrierType.UPANDIN, 100, 'analytic', 8.448206354250173),
    (BarrierType.UPANDIN, 110, 'analytic', 4.590969266108855),
    (BarrierType.UPANDIN, 90, 'fd-bs-barrier', 14.113898622657395),
    (BarrierType.UPANDIN, 90, 'fd-heston-barrier', 15.806560344445511)
])
def test_barrier_option_analytic_barrier_engine(my_barrier_option: BarrierOption,
                                                strike: float,
                                                barrier_type: BarrierType,
                                                my_market_data: MarketData,
                                                my_static_data: StaticData,
                                                model: str,
                                                expected: float) -> None:
    # Arrange
    my_barrier_option.strike = strike
    my_barrier_option.barrier_type = barrier_type
    # Act
    pv = price(my_barrier_option, my_market_data, my_static_data, model)

    # Assert
    assert pv == pytest.approx(expected, abs=1e-8)


def test_barrier_option_fd_black_scholes_rebate_engine(my_barrier_option: BarrierOption,
                                                       my_market_data: MarketData,
                                                       my_static_data: StaticData) -> None:
    # Act
    model = 'fd-bs-rebate'
    pv = price(my_barrier_option, my_market_data, my_static_data, model)

    # Assert
    assert pv == pytest.approx(2.9568061345268424, abs=1e-10)


def test_barrier_option_fd_heston_barrier_engine_constant_vol(my_barrier_option: BarrierOption,
                                                              my_market_data: MarketData,
                                                              my_static_data: StaticData) -> None:
    # Act
    model = 'fd-heston-barrier'
    # test if we retrieve the same price as BS
    my_market_data.data = [[my_market_data.underlying_black_scholes_volatilities[0]] * len(i) for i in
                           my_market_data.data]
    pv = price(my_barrier_option, my_market_data, my_static_data, model)

    # Assert
    assert pv == pytest.approx(14.114219673481117, abs=1e-8)



import pytest

from exotx import price
from exotx.data.marketdata import MarketData
from exotx.data.staticdata import StaticData
from exotx.enums.enums import PricingModel, NumericalMethod
from exotx.instruments.option_type import OptionType
from exotx.instruments.vanilla_option import VanillaOption
from exotx.utils.pricing_configuration import PricingConfiguration


# Arrange
@pytest.fixture
def my_vanilla_option() -> VanillaOption:
    strike = 90
    maturity = '2016-05-04'
    option_type = OptionType.CALL

    return VanillaOption(strike, maturity, option_type)


@pytest.fixture
def my_pricing_config() -> PricingConfiguration:
    model = PricingModel.BLACK_SCHOLES
    numerical_method = NumericalMethod.ANALYTIC
    compute_greeks = True

    return PricingConfiguration(model, numerical_method, compute_greeks=compute_greeks)


def test_price_compute_greeks_is_true(my_vanilla_option, my_market_data, my_static_data, my_pricing_config):
    # Act
    result = price(my_vanilla_option, my_market_data, my_static_data, my_pricing_config)

    # Assert
    assert 'price' in result
    assert 'delta' in result
    assert 'gamma' in result
    assert 'theta' in result


def test_price_compute_greeks_is_false(my_vanilla_option, my_market_data, my_static_data, my_pricing_config):
    # Arrange
    my_pricing_config.compute_greeks = False

    # Act
    result = price(my_vanilla_option, my_market_data, my_static_data, my_pricing_config)

    # Assert
    assert 'price' in result
    assert 'delta' not in result
    assert 'gamma' not in result
    assert 'theta' not in result


def test_price_invalid_model(my_vanilla_option, my_market_data, my_static_data, my_pricing_config):
    # Arrange
    my_pricing_config.model = 'invalid_model'

    # Act
    with pytest.raises(ValueError):
        price(my_vanilla_option, my_market_data, my_static_data, my_pricing_config)


@pytest.mark.parametrize(
    'strike, model, compute_greeks, expected_price, expected_delta, expected_gamma, expected_theta',
    [
        (90, PricingModel.BLACK_SCHOLES, True, 13.83328710, 0.77183751, 0.01609460, -7.01024983),
        (100, PricingModel.BLACK_SCHOLES, True, 7.84942762, 0.56837420, 0.02167605, -8.41931025),
        (110, PricingModel.BLACK_SCHOLES, True, 3.97951968, 0.36053753, 0.02089515, -7.65352494),
        (90, PricingModel.HESTON, False, 15.51921588, 0, 0, 0)
    ])
def test_price(my_vanilla_option: VanillaOption,
               my_market_data: MarketData,
               my_static_data: StaticData,
               my_pricing_config: PricingConfiguration,
               strike: float,
               model: str,
               compute_greeks: bool,
               expected_price: float,
               expected_delta: float,
               expected_gamma: float,
               expected_theta: float) -> None:
    # Arrange
    my_vanilla_option.strike = strike
    my_pricing_config.model = model
    my_pricing_config.compute_greeks = compute_greeks

    # Act
    result = price(my_vanilla_option, my_market_data, my_static_data, my_pricing_config, seed=125)

    # Assert
    assert result['price'] == pytest.approx(expected_price, abs=1e-4)
    if my_pricing_config.compute_greeks:
        assert result['delta'] == pytest.approx(expected_delta, abs=1e-8)
        assert result['gamma'] == pytest.approx(expected_gamma, abs=1e-8)
        assert result['theta'] == pytest.approx(expected_theta, abs=1e-8)



"""Modules for working with market and static data."""

from exotx.data.marketdata import MarketData
from exotx.data.staticdata import StaticData, StaticDataSchema

__all__ = [
    'MarketData',
    'StaticData',
    'StaticDataSchema'
]



from datetime import datetime

from exotx.data.marketdata import MarketData


def test_market_data_from_json():
    # Arrange
    my_json = {
        'reference_date': '2015-11-06',
        'underlying_spots': [100],
        'risk_free_rate': 0.01,
        'dividend_rate': 0,
        'underlying_black_scholes_volatilities': [0.2]
    }

    # Act
    my_market_data = MarketData.from_json(my_json)

    # Assert
    assert isinstance(my_market_data.reference_date, datetime)



from typing import Union, Tuple

from marshmallow import Schema, fields, post_load, validates_schema

from exotx.data.static.calendars import CalendarRegion, CalendarMarket, region_to_markets, available_regions


class Calendar:
    def __init__(self,
                 region: Union[CalendarRegion, str, None] = None,
                 market: Union[CalendarMarket, str, None] = None):
        self.region, self.market = Calendar.validate_inputs(region, market)

    @staticmethod
    def validate_inputs(region, market) -> Tuple[Union[CalendarRegion, None], Union[CalendarMarket, None]]:
        if region:
            if isinstance(region, str):
                try:
                    region = CalendarRegion[region]
                except KeyError as err:
                    raise ValueError(
                        f"Invalid region \'{region}\'. Available regions: {', '.join(available_regions)}") from err
            elif isinstance(region, CalendarRegion):
                pass
            else:
                raise TypeError(f"Invalid type for region \'{region.__repr__()}\'")
            if market:
                if isinstance(market, str):
                    try:
                        market = CalendarMarket[market]
                    except KeyError as err:
                        available_markets = [market.name for market in region_to_markets[region]]
                        raise ValueError(
                            f"Invalid market \'{market}\' for region \'{region.name}\'. Available markets: {', '.join(available_markets)}") from err
                Calendar.check_market_with_respect_to_region(region, market)
            else:
                # set default market
                market = CalendarMarket.Settlement
        else:
            if market:
                if isinstance(market, str):
                    raise ValueError(f"Missing region for market \'{market}\'")
                elif isinstance(market, CalendarMarket):
                    raise ValueError(f"Missing region for market \'{market.name}\'")
                else:
                    raise TypeError(f"Invalid type for market")
            else:
                # set default market
                region = None
                market = None

        return region, market

    @staticmethod
    def check_market_with_respect_to_region(region: CalendarRegion, market: CalendarMarket):
        if market not in region_to_markets[region]:
            available_markets = [market.name for market in region_to_markets[region]]
            raise ValueError(
                f"Invalid market \'{market.name}\' for region \'{region.name}\'. Available markets: {', '.join(available_markets)}")

    @staticmethod
    def get_default_calendar():
        return Calendar(region=None, market=None)

    @staticmethod
    def from_json(data: dict):
        schema = CalendarSchema()
        return schema.load(data)

    def to_json(self):
        schema = CalendarSchema()
        return schema.dump(self)


class CalendarRegionField(fields.Field):
    def _serialize(self, value: CalendarRegion, attr, obj, **kwargs) -> str:
        if value is not None:
            return value.name

    def _deserialize(self, value: str, attr, data, **kwargs) -> str:
        return value


class CalendarMarketField(fields.Field):
    def _serialize(self, value: CalendarMarket, attr, obj, **kwargs) -> str:
        if value is not None:
            return value.name

    def _deserialize(self, value: str, attr, data, **kwargs) -> str:
        return value


class CalendarSchema(Schema):
    region = CalendarRegionField(allow_none=True)
    market = CalendarMarketField(allow_none=True)

    @validates_schema
    def validate_region_and_market(self, data, **kwargs):
        if 'region' in data:
            if 'market' in data:
                _ = Calendar.validate_inputs(data['region'], data['market'])
            else:
                _ = Calendar.validate_inputs(data['region'], None)
        else:
            _ = Calendar.validate_inputs(None, data['market'])

    @post_load
    def make_calendar(self, data, **kwargs) -> Calendar:
        return Calendar(**data)



import enum

import QuantLib as ql


# based on https://quantlib-python-docs.readthedocs.io/en/latest/dates.html#conventions


# region Compounding
class Compounding(enum.Enum):
    Simple = 0
    Compounded = 1
    Continuous = 2
    SimpleThenCompounded = 3
    CompoundedThenSimple = 4


compoundings = {
    'Simple': ql.Simple,
    'Compounded': ql.Compounded,
    'Continuous': ql.Continuous,
    'SimpleThenCompounded': ql.SimpleThenCompounded,
    'CompoundedThenSimple': ql.CompoundedThenSimple
}


# endregion


# region Frequencies
class Frequencies(enum.Enum):
    NoFrequency = 0,
    Once = 1
    Annual = 2
    Semiannual = 3
    EveryFourthMonth = 4
    Quarterly = 5
    Bimonthly = 6
    Monthly = 7
    EveryFourthWeek = 8
    Biweekly = 9
    Weekly = 10
    Daily = 11


frequencies = {
    'NoFrequency': ql.NoFrequency,
    'Once': ql.Once,
    'Annual': ql.Annual,
    'Semiannual': ql.Semiannual,
    'EveryFourthMonth': ql.EveryFourthMonth,
    'Quarterly': ql.Quarterly,
    'Bimonthly': ql.Bimonthly,
    'Monthly': ql.Monthly,
    'EveryFourthWeek': ql.EveryFourthWeek,
    'Biweekly': ql.Biweekly,
    'Weekly': ql.Weekly,
    'Daily': ql.Daily
}


# endregion


# region Business day convention
class BusinessDayConvention(enum.Enum):
    Following = 0
    ModifiedFollowing = 1
    Preceding = 2
    ModifiedPreceding = 3
    Unadjusted = 4


business_day_conventions_to_ql = {
    BusinessDayConvention.Following: ql.Following,
    BusinessDayConvention.ModifiedFollowing: ql.ModifiedFollowing,
    BusinessDayConvention.Preceding: ql.Preceding,
    BusinessDayConvention.ModifiedPreceding: ql.ModifiedPreceding,
    BusinessDayConvention.Unadjusted: ql.Unadjusted
}
# endregion



import json
from typing import Union

import QuantLib as ql
from marshmallow import Schema, fields, post_load, ValidationError

from exotx.data.static.calendar import Calendar, CalendarSchema
from exotx.data.static.calendars import calendars_to_ql_calendars
from exotx.data.static.conventions import business_day_conventions_to_ql, BusinessDayConvention
from exotx.data.static.daycounters import day_counters_to_ql, DayCounter


class StaticData:
    def __init__(self,
                 day_counter: Union[DayCounter, str, None] = None,
                 business_day_convention: Union[BusinessDayConvention, str, None] = None,
                 calendar: Union[Calendar, dict, None] = None) -> None:
        # day counter
        self._set_day_counter(day_counter)

        # business day convention
        self._set_business_day_convention(business_day_convention)

        # calendar
        self._set_calendar(calendar)

    # region Day counter
    @staticmethod
    def get_default_day_counter() -> DayCounter:
        return DayCounter.Actual360

    def _set_day_counter(self, day_counter: Union[DayCounter, str, None]) -> None:
        if isinstance(day_counter, DayCounter):
            self.day_counter = day_counter
        elif isinstance(day_counter, str):
            try:
                self.day_counter = DayCounter[day_counter]
            except KeyError as err:
                raise ValueError(f"Invalid day counter \'{day_counter}\'") from err
        elif day_counter is None:
            self.day_counter: DayCounter = self.get_default_day_counter()
        else:
            raise ValueError(f"Invalid day counter type {type(day_counter)}")

    def get_ql_day_counter(self) -> ql.DayCounter:
        return day_counters_to_ql[self.day_counter]

    @staticmethod
    def get_ql_default_day_counter():
        return day_counters_to_ql[StaticData.get_default_day_counter()]

    # endregion

    # region Calendar
    @staticmethod
    def get_default_calendar() -> Calendar:
        return Calendar.get_default_calendar()

    @staticmethod
    def get_ql_default_calendar() -> ql.Calendar:
        return ql.TARGET()

    def get_ql_calendar(self) -> ql.Calendar:
        if self.calendar.region:
            # TODO: Make sure market is defined as CalendarMarket.Settlement here if market was None
            return calendars_to_ql_calendars[self.calendar.region][self.calendar.market]
        else:
            return self.get_ql_default_calendar()

    def _set_calendar(self, value: Union[Calendar, dict, None]) -> None:
        if isinstance(value, Calendar):
            self.calendar = value
        elif isinstance(value, dict):
            if 'region' in value.keys():
                region = value['region']
                assert region in list(calendars_to_ql_calendars.keys()), f"Invalid region \'{region}\'"
                if 'market' in value.keys():
                    market = value['market']
                    assert market in list(calendars_to_ql_calendars[
                                              region].keys()), f"Invalid market \'{market}\' for region \'{region}\'"
                    self.calendar = Calendar(region=region, market=market)
                else:
                    # set calendar with default market
                    self.calendar = Calendar(region=region, market=None)
            else:
                if 'market' in value.keys():
                    raise Exception(f"Missing calendar for market \'{value['market']}\'")
                else:
                    # set default calendar
                    self.calendar = self.get_default_calendar()
        elif value is None:
            self.calendar = self.get_default_calendar()
        else:
            raise ValueError(f"Invalid calendar type {type(value)}")

    # endregion

    # region Business day convention
    @staticmethod
    def get_default_business_day_convention() -> BusinessDayConvention:
        return BusinessDayConvention.ModifiedFollowing

    def get_default_ql_business_day_convention(self) -> int:
        return business_day_conventions_to_ql[self.business_day_convention]

    def _set_business_day_convention(self, value: Union[BusinessDayConvention, str, None]) -> None:
        if isinstance(value, BusinessDayConvention):
            self.business_day_convention = value
        elif isinstance(value, str):
            try:
                self.business_day_convention = BusinessDayConvention[value]
            except KeyError as err:
                raise ValueError(f"Invalid business day convention \'{value}\'") from err
        elif value is None:
            self.business_day_convention = self.get_default_business_day_convention()
        else:
            raise ValueError(f"Invalid business day convention type {type(value)}")

    # endregion

    @classmethod
    def from_json(cls, data: dict):
        schema = StaticDataSchema()
        return schema.load(data)

    def to_json(self, format_type: str = "dict"):
        schema = StaticDataSchema()
        my_json = schema.dump(self)
        if format_type == "dict":
            return my_json
        elif format_type == "str":
            return json.dumps(my_json)
        else:
            raise NotImplementedError(f"Invalid format type {format_type} when dumping")


# region Schema

class BusinessDayConventionField(fields.Field):
    def _serialize(self, value: BusinessDayConvention, attr, obj, **kwargs) -> str:
        return value.name

    def _deserialize(self, value: str, attr, data, **kwargs) -> BusinessDayConvention:
        try:
            if value:
                return BusinessDayConvention[value]
            else:
                return StaticData.get_default_business_day_convention()
        except KeyError as error:
            raise ValidationError(f"Invalid business day convention \'{value}\'") from error


class DayCounterField(fields.Field):
    def _serialize(self, value: DayCounter, attr, obj, **kwargs) -> str:
        return value.name

    def _deserialize(self, value: str, attr, data, **kwargs) -> DayCounter:
        try:
            if value:
                return DayCounter[value]
            else:
                return StaticData.get_default_day_counter()
        except KeyError as error:
            raise ValidationError(f"Invalid day counter \'{value}\'") from error


class StaticDataSchema(Schema):
    day_counter = DayCounterField(allow_none=True)
    business_day_convention = BusinessDayConventionField(allow_none=True)
    calendar = fields.Nested(CalendarSchema(), allow_none=True)

    @post_load
    def make_static_data(self, data, **kwargs) -> StaticData:
        return StaticData(**data)

# endregion



import enum

import QuantLib as ql


# based on https://quantlib-python-docs.readthedocs.io/en/latest/dates.html#calendar


class CalendarRegion(enum.Enum):
    Argentina = 0
    Australia = 1
    BespokeCalendar = 2
    Brazil = 3
    Canada = 4
    China = 5
    CzechRepublic = 6
    Denmark = 7
    Finland = 8
    France = 9
    Germany = 10
    HongKong = 11
    Hungary = 12
    Iceland = 13
    India = 14
    Indonesia = 15
    Israel = 16
    Italy = 17
    Japan = 18
    Mexico = 19
    NewZealand = 20
    Norway = 21
    NullCalendar = 22
    Poland = 23
    Romania = 24
    Russia = 25
    SaudiArabia = 26
    Singapore = 27
    Slovakia = 28
    SouthAfrica = 29
    SouthKorea = 30
    Sweden = 31
    Switzerland = 32
    Taiwan = 33
    TARGET = 34
    Thailand = 35
    Turkey = 36
    Ukraine = 37
    UnitedKingdom = 38
    UnitedStates = 39
    WeekendsOnly = 40


class CalendarMarket(enum.Enum):
    Settlement = 0
    Merval = 1
    Exchange = 2
    TSX = 3
    IB = 4
    SSE = 5
    PSE = 6
    Eurex = 7
    FrankfurtStockExchange = 8
    Xetra = 9
    HKEx = 10
    ICEX = 11
    NSE = 12
    BEJ = 13
    JSX = 14
    TASE = 15
    BMV = 16
    MOEX = 17
    Tadawul = 18
    SGX = 19
    BSSE = 20
    KRX = 21
    TSEC = 22
    USE = 23
    Metals = 24
    FederalReserve = 25
    GovernmentBond = 26
    LiborImpact = 27
    NERC = 28
    NYSE = 29


region_to_markets = {
    CalendarRegion.Argentina: [
        CalendarMarket.Settlement,
        CalendarMarket.Merval
    ],
    CalendarRegion.Australia: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.BespokeCalendar: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Brazil: [
        CalendarMarket.Settlement,
        CalendarMarket.Exchange
    ],
    CalendarRegion.Canada: [
        CalendarMarket.Settlement,
        CalendarMarket.TSX
    ],
    CalendarRegion.China: [
        CalendarMarket.Settlement,
        CalendarMarket.IB,
        CalendarMarket.SSE
    ],
    CalendarRegion.CzechRepublic: [
        CalendarMarket.Settlement,
        CalendarMarket.PSE
    ],
    CalendarRegion.Denmark: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Finland: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.France: [
        CalendarMarket.Settlement,
        CalendarMarket.Exchange
    ],
    CalendarRegion.Germany: [
        CalendarMarket.Settlement,
        CalendarMarket.Eurex,
        CalendarMarket.FrankfurtStockExchange,
        CalendarMarket.Xetra
    ],
    CalendarRegion.HongKong: [
        CalendarMarket.Settlement,
        CalendarMarket.HKEx
    ],
    CalendarRegion.Hungary: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Iceland: [
        CalendarMarket.Settlement,
        CalendarMarket.ICEX
    ],
    CalendarRegion.India: [
        CalendarMarket.Settlement,
        CalendarMarket.NSE
    ],
    CalendarRegion.Indonesia: [
        CalendarMarket.Settlement,
        CalendarMarket.BEJ,
        CalendarMarket.JSX
    ],
    CalendarRegion.Israel: [
        CalendarMarket.Settlement,
        CalendarMarket.TASE
    ],
    CalendarRegion.Italy: [
        CalendarMarket.Settlement,
        CalendarMarket.Exchange
    ],
    CalendarRegion.Japan: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Mexico: [
        CalendarMarket.Settlement,
        CalendarMarket.BMV
    ],
    CalendarRegion.NewZealand: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Norway: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.NullCalendar: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Poland: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Romania: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Russia: [
        CalendarMarket.Settlement,
        CalendarMarket.MOEX
    ],
    CalendarRegion.SaudiArabia: [
        CalendarMarket.Settlement,
        CalendarMarket.Tadawul
    ],
    CalendarRegion.Singapore: [
        CalendarMarket.Settlement,
        CalendarMarket.SGX
    ],
    CalendarRegion.Slovakia: [
        CalendarMarket.Settlement,
        CalendarMarket.BSSE
    ],
    CalendarRegion.SouthAfrica: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.SouthKorea: [
        CalendarMarket.Settlement,
        CalendarMarket.KRX
    ],
    CalendarRegion.Sweden: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Switzerland: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Taiwan: [
        CalendarMarket.Settlement,
        CalendarMarket.TSEC
    ],
    CalendarRegion.TARGET: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Thailand: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Turkey: [
        CalendarMarket.Settlement
    ],
    CalendarRegion.Ukraine: [
        CalendarMarket.Settlement,
        CalendarMarket.USE
    ],
    CalendarRegion.UnitedKingdom: [
        CalendarMarket.Settlement,
        CalendarMarket.Exchange,
        CalendarMarket.Metals
    ],
    CalendarRegion.UnitedStates: [
        CalendarMarket.Settlement,
        CalendarMarket.FederalReserve,
        CalendarMarket.GovernmentBond,
        CalendarMarket.LiborImpact,
        CalendarMarket.NERC,
        CalendarMarket.NYSE
    ],
    CalendarRegion.WeekendsOnly: [
        CalendarMarket.Settlement
    ]
}

available_regions = [region.name for region in CalendarRegion]

calendars_to_ql_calendars = {
    CalendarRegion.Argentina: {
        CalendarMarket.Merval: ql.Argentina(ql.Argentina.Merval),
        CalendarMarket.Settlement: ql.Argentina()
    },
    CalendarRegion.Australia: {
        CalendarMarket.Settlement: ql.Australia()
    },
    CalendarRegion.BespokeCalendar: {
        CalendarMarket.Settlement: ql.BespokeCalendar('BespokeCalendar')
    },
    CalendarRegion.Brazil: {
        CalendarMarket.Exchange: ql.Brazil(ql.Brazil.Exchange),
        CalendarMarket.Settlement: ql.Brazil(ql.Brazil.Settlement)
    },
    CalendarRegion.Canada: {
        CalendarMarket.Settlement: ql.Canada(ql.Canada.Settlement),
        CalendarMarket.TSX: ql.Canada(ql.Canada.TSX)
    },
    CalendarRegion.China: {
        CalendarMarket.IB: ql.China(ql.China.IB),
        CalendarMarket.SSE: ql.China(ql.China.SSE),
        CalendarMarket.Settlement: ql.China()
    },
    CalendarRegion.CzechRepublic: {
        CalendarMarket.PSE: ql.CzechRepublic(ql.CzechRepublic.PSE),
        CalendarMarket.Settlement: ql.CzechRepublic()
    },
    CalendarRegion.Denmark: {
        CalendarMarket.Settlement: ql.Denmark()
    },
    CalendarRegion.Finland: {
        CalendarMarket.Settlement: ql.Finland()
    },
    CalendarRegion.France: {
        CalendarMarket.Exchange: ql.France(ql.France.Exchange),
        CalendarMarket.Settlement: ql.France(ql.France.Settlement)
    },
    CalendarRegion.Germany: {
        CalendarMarket.Eurex: ql.Germany(ql.Germany.Eurex),
        CalendarMarket.FrankfurtStockExchange: ql.Germany(ql.Germany.FrankfurtStockExchange),
        CalendarMarket.Settlement: ql.Germany(ql.Germany.Settlement),
        CalendarMarket.Xetra: ql.Germany(ql.Germany.Xetra)
    },
    CalendarRegion.HongKong: {
        CalendarMarket.HKEx: ql.HongKong(ql.HongKong.HKEx),
        CalendarMarket.Settlement: ql.HongKong()
    },
    CalendarRegion.Hungary: {
        CalendarMarket.Settlement: ql.Hungary()
    },
    CalendarRegion.Iceland: {
        CalendarMarket.ICEX: ql.Iceland(ql.Iceland.ICEX),
        CalendarMarket.Settlement: ql.Iceland()
    },
    CalendarRegion.India: {
        CalendarMarket.NSE: ql.India(ql.India.NSE),
        CalendarMarket.Settlement: ql.India()
    },
    CalendarRegion.Indonesia: {
        CalendarMarket.BEJ: ql.Indonesia(ql.Indonesia.BEJ),
        CalendarMarket.JSX: ql.Indonesia(ql.Indonesia.JSX),
        CalendarMarket.Settlement: ql.Indonesia()
    },
    CalendarRegion.Israel: {
        CalendarMarket.Settlement: ql.Israel(ql.Israel.Settlement),
        CalendarMarket.TASE: ql.Israel(ql.Israel.TASE)
    },
    CalendarRegion.Italy: {
        CalendarMarket.Exchange: ql.Italy(ql.Italy.Exchange),
        CalendarMarket.Settlement: ql.Italy(ql.Italy.Settlement)
    },
    CalendarRegion.Japan: {
        CalendarMarket.Settlement: ql.Japan()
    },
    CalendarRegion.Mexico: {
        CalendarMarket.BMV: ql.Mexico(ql.Mexico.BMV),
        CalendarMarket.Settlement: ql.Mexico()
    },
    CalendarRegion.NewZealand: {
        CalendarMarket.Settlement: ql.NewZealand()
    },
    CalendarRegion.Norway: {
        CalendarMarket.Settlement: ql.Norway()
    },
    CalendarRegion.NullCalendar: {
        CalendarMarket.Settlement: ql.NullCalendar()
    },
    CalendarRegion.Poland: {
        CalendarMarket.Settlement: ql.Poland()
    },
    CalendarRegion.Romania: {
        CalendarMarket.Settlement: ql.Romania()
    },
    CalendarRegion.Russia: {
        CalendarMarket.MOEX: ql.Russia(ql.Russia.MOEX),
        CalendarMarket.Settlement: ql.Russia(ql.Russia.Settlement)
    },
    CalendarRegion.SaudiArabia: {
        CalendarMarket.Tadawul: ql.SaudiArabia(ql.SaudiArabia.Tadawul),
        CalendarMarket.Settlement: ql.SaudiArabia()
    },
    CalendarRegion.Singapore: {
        CalendarMarket.SGX: ql.Singapore(ql.Singapore.SGX),
        CalendarMarket.Settlement: ql.Singapore()
    },
    CalendarRegion.Slovakia: {
        CalendarMarket.BSSE: ql.Slovakia(ql.Slovakia.BSSE),
        CalendarMarket.Settlement: ql.Slovakia()
    },
    CalendarRegion.SouthAfrica: {
        CalendarMarket.Settlement: ql.SouthAfrica()
    },
    CalendarRegion.SouthKorea: {
        CalendarMarket.KRX: ql.SouthKorea(ql.SouthKorea.KRX),
        CalendarMarket.Settlement: ql.SouthKorea(ql.SouthKorea.Settlement)
    },
    CalendarRegion.Sweden: {
        CalendarMarket.Settlement: ql.Sweden()
    },
    CalendarRegion.Switzerland: {
        CalendarMarket.Settlement: ql.Switzerland()
    },
    CalendarRegion.Taiwan: {
        CalendarMarket.TSEC: ql.Taiwan(ql.Taiwan.TSEC),
        CalendarMarket.Settlement: ql.Taiwan()
    },
    CalendarRegion.TARGET: {
        CalendarMarket.Settlement: ql.TARGET()
    },
    CalendarRegion.Thailand: {
        CalendarMarket.Settlement: ql.Thailand()
    },
    CalendarRegion.Turkey: {
        CalendarMarket.Settlement: ql.Turkey()
    },
    CalendarRegion.Ukraine: {
        CalendarMarket.USE: ql.Ukraine(ql.Ukraine.USE),
        CalendarMarket.Settlement: ql.Ukraine()
    },
    CalendarRegion.UnitedKingdom: {
        CalendarMarket.Exchange: ql.UnitedKingdom(ql.UnitedKingdom.Exchange),
        CalendarMarket.Metals: ql.UnitedKingdom(ql.UnitedKingdom.Metals),
        CalendarMarket.Settlement: ql.UnitedKingdom(ql.UnitedKingdom.Settlement)
    },
    CalendarRegion.UnitedStates: {
        CalendarMarket.FederalReserve: ql.UnitedStates(ql.UnitedStates.FederalReserve),
        CalendarMarket.GovernmentBond: ql.UnitedStates(ql.UnitedStates.GovernmentBond),
        CalendarMarket.LiborImpact: ql.UnitedStates(ql.UnitedStates.LiborImpact),
        CalendarMarket.NERC: ql.UnitedStates(ql.UnitedStates.NERC),
        CalendarMarket.NYSE: ql.UnitedStates(ql.UnitedStates.NYSE),
        CalendarMarket.Settlement: ql.UnitedStates(ql.UnitedStates.Settlement)
    },
    CalendarRegion.WeekendsOnly: {
        CalendarMarket.Settlement: ql.WeekendsOnly()
    }
}



from datetime import datetime
from typing import Union, List

import QuantLib as ql
from marshmallow import Schema, fields, post_load

from exotx.enums.enums import PricingModel, NumericalMethod
from exotx.helpers.dates import convert_maturity_to_ql_date
from exotx.instruments.average_calculation import AverageCalculation, convert_average_calculation, \
    AverageCalculationField
from exotx.instruments.average_convention import AverageConvention, convert_average_convention, AverageConventionField
from exotx.instruments.average_type import AverageTypeField, AverageType, convert_average_type_to_ql
from exotx.instruments.instrument import Instrument
from exotx.instruments.option_type import convert_option_type_to_ql, OptionType, OptionTypeField
from exotx.models.blackscholesmodel import BlackScholesModel
from exotx.utils.pricing_configuration import PricingConfiguration


class AsianOption(Instrument):
    def __init__(self,
                 strike: float, maturity: Union[str, datetime], option_type: Union[str, OptionType],
                 average_type: Union[str, AverageType], average_calculation: Union[str, AverageCalculation],
                 average_convention: Union[str, AverageConvention], arithmetic_running_accumulator: float = 0.0,
                 geometric_running_accumulator: float = 1.0, past_fixings: int = 0,
                 future_fixing_dates: List[datetime] = None):
        assert strike >= 0, "Invalid strike: cannot be negative"
        self.strike = strike
        self.maturity = convert_maturity_to_ql_date(maturity)
        self.option_type = convert_option_type_to_ql(option_type)
        self.average_type = convert_average_type_to_ql(average_type)
        self.average_calculation: AverageCalculation = convert_average_calculation(average_calculation)
        self.average_convention: AverageConvention = convert_average_convention(average_convention)
        self.arithmetic_running_accumulator = arithmetic_running_accumulator
        self.geometric_running_accumulator = geometric_running_accumulator
        self.past_fixings = past_fixings
        self.future_fixing_dates = None if not future_fixing_dates else [ql.Date().from_date(future_fixing_date)
                                                                         for future_fixing_date in future_fixing_dates]

    def price(self, market_data, static_data, pricing_config: PricingConfiguration, seed: int = 1) -> dict:
        reference_date: ql.Date = market_data.get_ql_reference_date()
        ql.Settings.instance().evaluationDate = reference_date

        # check future fixing dates
        if self.future_fixing_dates:
            for future_fixing_date in self.future_fixing_dates:
                assert future_fixing_date >= reference_date, f"Invalid future fixing date {future_fixing_date}"

        # create the product
        ql_payoff = ql.PlainVanillaPayoff(self.option_type, self.strike)
        ql_exercise = ql.EuropeanExercise(self.maturity)
        if self.average_calculation == AverageCalculation.CONTINUOUS:
            ql_option = ql.ContinuousAveragingAsianOption(self.average_type, ql_payoff, ql_exercise)
        elif self.average_calculation == AverageCalculation.DISCRETE:
            if self.average_type == ql.Average().Arithmetic:
                ql_option = ql.DiscreteAveragingAsianOption(self.average_type,
                                                            self.arithmetic_running_accumulator,
                                                            self.past_fixings, self.future_fixing_dates, ql_payoff,
                                                            ql_exercise)
            elif self.average_type == ql.Average().Geometric:
                ql_option = ql.DiscreteAveragingAsianOption(self.average_type, self.geometric_running_accumulator,
                                                            self.past_fixings, self.future_fixing_dates, ql_payoff,
                                                            ql_exercise)
            else:
                raise ValueError(f"Invalid average type \"{self.average_type}\"")
        else:
            raise ValueError(f"Invalid average calculation \"{self.average_calculation}\"")

        # set the pricing engine
        ql_engine = self._get_ql_pricing_engine(market_data, static_data, pricing_config, seed)
        ql_option.setPricingEngine(ql_engine)

        # price
        price = ql_option.NPV()
        if pricing_config.compute_greeks and pricing_config.model != PricingModel.HESTON:
            delta = ql_option.delta()
            gamma = ql_option.gamma()
            theta = ql_option.theta()
            return {'price': price, 'delta': delta, 'gamma': gamma, 'theta': theta}
        else:
            return {'price': price}

    def _get_ql_pricing_engine(self, market_data, static_data, pricing_config: PricingConfiguration, seed: int):
        if self.average_calculation == AverageCalculation.DISCRETE:
            if self.average_type == ql.Average().Geometric:
                if pricing_config.numerical_method == NumericalMethod.ANALYTIC:
                    # TODO: filter on pricing_config.pricing_model, here we assume black-scholes only
                    bs_model = BlackScholesModel(market_data, static_data)
                    process = bs_model.setup()
                    if self.average_convention == AverageConvention.PRICE:
                        return ql.AnalyticDiscreteGeometricAveragePriceAsianEngine(process)
                    elif self.average_convention == AverageConvention.STRIKE:
                        return ql.AnalyticDiscreteGeometricAverageStrikeAsianEngine(process)
                    else:
                        raise ValueError(f"Invalid average convention \"{self.average_convention}\"")
                elif pricing_config.numerical_method == NumericalMethod.MC:
                    # TODO: filter on pricing_config.pricing_model, here we assume black-scholes only
                    bs_model = BlackScholesModel(market_data, static_data)
                    process = bs_model.setup()
                    random_number_generator = str(pricing_config.random_number_generator)
                    if self.average_convention == AverageConvention.PRICE:
                        return ql.MCDiscreteGeometricAPEngine(process, random_number_generator)
                    elif self.average_convention == AverageConvention.STRIKE:
                        return ValueError(
                            f"No corresponding engine for asian option for numerical method {pricing_config.numerical_method}, "
                            f"average type {self.average_type}, average calculation {self.average_calculation}, and average convention {self.average_convention}")
                    else:
                        raise ValueError(f"Invalid average convention \"{self.average_convention}\"")
                else:
                    raise ValueError(
                        f"No engine for asian option with numerical method {pricing_config.numerical_method}"
                        f"with average calculation {self.average_calculation} and average type {self.average_type}")
            elif self.average_type == ql.Average().Arithmetic:
                if pricing_config.numerical_method == NumericalMethod.MC:
                    bs_model = BlackScholesModel(market_data, static_data)
                    process = bs_model.setup()
                    random_number_generator = str(pricing_config.random_number_generator)
                    if self.average_convention == AverageConvention.PRICE:
                        return ql.MCDiscreteArithmeticAPEngine(process, random_number_generator)
                    elif self.average_convention == AverageConvention.STRIKE:
                        return ql.MCDiscreteArithmeticASEngine(process, random_number_generator)
                    else:
                        raise ValueError(f"Invalid average convention \"{self.average_convention}\"")
            else:
                raise ValueError(f"Invalid average type {self.average_type}")
        elif self.average_calculation == AverageCalculation.CONTINUOUS:
            if self.average_type == ql.Average().Geometric:
                if pricing_config.numerical_method == NumericalMethod.ANALYTIC:
                    # TODO: filter on pricing_config.pricing_model, here we assume black-scholes only
                    bs_model = BlackScholesModel(market_data, static_data)
                    process = bs_model.setup()
                    return ql.AnalyticContinuousGeometricAveragePriceAsianEngine(process)
                else:
                    raise ValueError(
                        f"No engine for asian option with numerical method {pricing_config.numerical_method}"
                        f"with average calculation {self.average_calculation} and average type {self.average_type}")
            else:
                raise ValueError("No engine for asian option")
        else:
            raise ValueError(f"Invalid average calculation \"{self.average_calculation}\"")

    # region serialization/deserialization
    def to_json(self):
        schema = AsianOptionSchema()
        return schema.dump(self)

    @classmethod
    def from_json(cls, json_data):
        schema = AsianOptionSchema()
        data = schema.load(json_data)
        return cls(**data)
    # endregion


# region Schema
class AsianOptionSchema(Schema):
    strike = fields.Float()
    maturity = fields.Date(format="%Y-%m-%d")
    option_type = OptionTypeField()
    average_type = AverageTypeField()
    average_calculation = AverageCalculationField()
    average_convention = AverageConventionField()

    @post_load
    def make_asian_option(self, data, **kwargs) -> AsianOption:
        return AsianOption(**data)
# endregion



"""Modules for working with exotic options.

Exotic options can refer to auto-callables, barrier options, etc."""

from exotx.instruments.autocallable import Autocallable
from exotx.instruments.barrier_option import BarrierOption
from exotx.instruments.instrument import price

__all__ = [
    'price',
    'Autocallable',
    'BarrierOption'
]



import pytest

from exotx import price
from exotx.data.marketdata import MarketData
from exotx.data.staticdata import StaticData
from exotx.instruments.autocallable import Autocallable


# Arrange
@pytest.fixture
def my_market_data() -> MarketData:
    reference_date = '2015-11-06'
    underlying_spots = [100.0]
    underlying_black_scholes_volatilities = [0.2]
    risk_free_rate = 0.01
    dividend_rate = 0.0

    return MarketData(reference_date=reference_date,
                      underlying_spots=underlying_spots,
                      risk_free_rate=risk_free_rate,
                      dividend_rate=dividend_rate,
                      underlying_black_scholes_volatilities=underlying_black_scholes_volatilities)


@pytest.fixture
def my_autocallable() -> Autocallable:
    notional = 100
    strike = 100
    autocall_barrier_level = 1.0
    coupon = 0.03
    coupon_barrier_level = 0.75
    protection_barrier_level = 0.75

    return Autocallable(notional, strike, autocall_barrier_level, coupon, coupon_barrier_level,
                        protection_barrier_level)


def test_autocallable_black_scholes_price(my_autocallable: Autocallable,
                                          my_market_data: MarketData,
                                          my_static_data: StaticData) -> None:
    # Act
    seed = 125
    model = 'black-scholes'
    pv = price(my_autocallable, my_market_data, my_static_data, model, seed)

    # Assert
    assert pv == pytest.approx(96.08517973497098, abs=1e-10)



import pytest

from exotx import price
from exotx.data.marketdata import MarketData
from exotx.data.staticdata import StaticData
from exotx.enums.enums import PricingModel, NumericalMethod
from exotx.instruments.basket_option import BasketOption
from exotx.instruments.basket_type import BasketType
from exotx.instruments.option_type import OptionType
from exotx.utils.pricing_configuration import PricingConfiguration


# Arrange
@pytest.fixture()
def my_static_data() -> StaticData:
    my_json = {
        'day_counter': 'Actual360',
        'business_day_convention': 'ModifiedFollowing'
    }
    return StaticData.from_json(my_json)


@pytest.fixture()
def my_market_data() -> MarketData:
    my_json = {
        'reference_date': '2015-11-06',
        'underlying_spots': [80, 90, 100],
        'risk_free_rate': 0.05,
        'dividend_rate': -0.03,
        'underlying_black_scholes_volatilities': [0.20, 0.25, 0.3],
        'correlation_matrix': [
            [1.0, 0.5, 0.6],
            [0.5, 1.0, 0.7],
            [0.6, 0.7, 1.0]
        ]
    }

    return MarketData.from_json(my_json)


@pytest.fixture
def my_basket_option() -> BasketOption:
    strike = 85
    maturity = '2016-02-04'  # reference_date + 90 days
    exercise = 'european'
    option_type = OptionType.PUT
    # option_type = 'call'
    basket_type = BasketType.MINBASKET

    return BasketOption(strike, maturity, option_type, basket_type)


@pytest.fixture
def my_pricing_config() -> PricingConfiguration:
    model = PricingModel.BLACK_SCHOLES
    numerical_method = NumericalMethod.ANALYTIC

    return PricingConfiguration(model, numerical_method)


@pytest.mark.parametrize(
    'expected_price',
    [
        5.7207
    ])
def test_price(my_basket_option: BasketOption,
               my_market_data: MarketData,
               my_static_data: StaticData,
               my_pricing_config: PricingConfiguration,
               expected_price: float) -> None:
    # Arrange
    seed = 42

    # Act
    result = price(my_basket_option, my_market_data, my_static_data, my_pricing_config, seed)

    # Assert
    assert result['price'] == pytest.approx(expected_price, abs=1e-4)



from typing import Optional, Union

import QuantLib as ql
import pytest

from exotx.data import StaticData
from exotx.data.static import Calendar, BusinessDayConvention, DayCounter


# Arrange
def my_default_day_counter_type() -> type:
    my_default_day_counter = StaticData.get_default_day_counter()
    return type(my_default_day_counter)


def my_default_calendar_type() -> type:
    my_default_calendar = StaticData.get_default_calendar()
    return type(my_default_calendar)


def my_default_business_day_convention_type() -> type:
    my_default_business_day_convention = StaticData.get_default_business_day_convention()
    return type(my_default_business_day_convention)


# region DayCounter
@pytest.mark.parametrize('day_counter, expected_class_type, expected_ql_class_type', [
    ('Actual360', DayCounter, ql.Actual360),
    ('SimpleDayCounter', DayCounter, ql.SimpleDayCounter),
    ('Thirty360', DayCounter, ql.Thirty360),
    ('Actual365Fixed', DayCounter, ql.Actual365Fixed),
    ('Actual365FixedCanadian', DayCounter, ql.Actual365Fixed),
    ('Actual365FixedNoLeap', DayCounter, ql.Actual365Fixed),
    ('ActualActual', DayCounter, ql.ActualActual),
    ('Business252', DayCounter, ql.Business252),
    (None, DayCounter, ql.Actual360)  # default behavior
])
def test_static_data_day_counter(day_counter: Optional[str],
                                 expected_class_type: type,
                                 expected_ql_class_type: type) -> None:
    # Act
    my_static_data = StaticData(day_counter=day_counter)

    # Assert
    assert isinstance(my_static_data.day_counter, expected_class_type)
    assert isinstance(my_static_data.get_ql_day_counter(), expected_ql_class_type)

    # default settings
    assert isinstance(my_static_data.calendar, my_default_calendar_type())
    assert isinstance(my_static_data.business_day_convention, my_default_business_day_convention_type())


@pytest.mark.parametrize('day_counter', [
    'actual360',  # invalid day counter
])
def test_static_data_invalid_day_counter(day_counter: Optional[str]) -> None:
    # Act
    with pytest.raises(Exception) as e:
        _ = StaticData(day_counter=day_counter)

    # Assert
    assert str(e.value) == f'Invalid day counter \'{day_counter}\''


# endregion DayCounter

# region Calendar
@pytest.mark.parametrize('region, market, expected_class_type, expected_ql_class_type', [
    ('Argentina', 'Merval', Calendar, ql.Argentina),
    ('Argentina', 'Settlement', Calendar, ql.Argentina),
    ('Argentina', None, Calendar, ql.Argentina),
    ('Australia', 'Settlement', Calendar, ql.Australia),
    ('Australia', None, Calendar, ql.Australia),
    ('BespokeCalendar', 'Settlement', Calendar, ql.BespokeCalendar),
    ('BespokeCalendar', None, Calendar, ql.BespokeCalendar),
    ('Brazil', 'Exchange', Calendar, ql.Brazil),
    ('Brazil', 'Settlement', Calendar, ql.Brazil),
    ('Brazil', None, Calendar, ql.Brazil),
    ('Canada', 'Settlement', Calendar, ql.Canada),
    ('Canada', 'TSX', Calendar, ql.Canada),
    ('Canada', None, Calendar, ql.Canada),
    ('China', 'IB', Calendar, ql.China),
    ('China', 'SSE', Calendar, ql.China),
    ('China', 'Settlement', Calendar, ql.China),
    ('China', None, Calendar, ql.China),
    ('CzechRepublic', 'PSE', Calendar, ql.CzechRepublic),
    ('CzechRepublic', 'Settlement', Calendar, ql.CzechRepublic),
    ('CzechRepublic', None, Calendar, ql.CzechRepublic),
    ('Denmark', 'Settlement', Calendar, ql.Denmark),
    ('Finland', 'Settlement', Calendar, ql.Finland),
    ('France', 'Exchange', Calendar, ql.France),
    ('France', 'Settlement', Calendar, ql.France),
    ('Germany', 'Eurex', Calendar, ql.Germany),
    ('Germany', 'FrankfurtStockExchange', Calendar, ql.Germany),
    ('Germany', 'Settlement', Calendar, ql.Germany),
    ('Germany', 'Xetra', Calendar, ql.Germany),
    ('Germany', None, Calendar, ql.Germany),
    ('HongKong', 'HKEx', Calendar, ql.HongKong),
    ('HongKong', 'Settlement', Calendar, ql.HongKong),
    ('HongKong', None, Calendar, ql.HongKong),
    ('Hungary', 'Settlement', Calendar, ql.Hungary),
    ('Hungary', None, Calendar, ql.Hungary),
    ('Iceland', 'ICEX', Calendar, ql.Iceland),
    ('Iceland', 'Settlement', Calendar, ql.Iceland),
    ('Iceland', None, Calendar, ql.Iceland),
    ('India', 'NSE', Calendar, ql.India),
    ('India', 'Settlement', Calendar, ql.India),
    ('India', None, Calendar, ql.India),
    ('Indonesia', 'BEJ', Calendar, ql.Indonesia),
    ('Indonesia', 'JSX', Calendar, ql.Indonesia),
    ('Indonesia', 'Settlement', Calendar, ql.Indonesia),
    ('Indonesia', None, Calendar, ql.Indonesia),
    ('Israel', 'Settlement', Calendar, ql.Israel),
    ('Israel', 'TASE', Calendar, ql.Israel),
    ('Israel', None, Calendar, ql.Israel),
    ('Italy', 'Exchange', Calendar, ql.Italy),
    ('Italy', 'Settlement', Calendar, ql.Italy),
    ('Italy', None, Calendar, ql.Italy),
    ('Japan', 'Settlement', Calendar, ql.Japan),
    ('Japan', None, Calendar, ql.Japan),
    ('Mexico', 'BMV', Calendar, ql.Mexico),
    ('Mexico', None, Calendar, ql.Mexico),
    ('NewZealand', 'Settlement', Calendar, ql.NewZealand),
    ('NewZealand', None, Calendar, ql.NewZealand),
    ('Norway', 'Settlement', Calendar, ql.Norway),
    ('Norway', None, Calendar, ql.Norway),
    ('NullCalendar', 'Settlement', Calendar, ql.NullCalendar),
    ('NullCalendar', None, Calendar, ql.NullCalendar),
    ('Poland', 'Settlement', Calendar, ql.Poland),
    ('Poland', None, Calendar, ql.Poland),
    ('Romania', 'Settlement', Calendar, ql.Romania),
    ('Romania', None, Calendar, ql.Romania),
    ('Russia', 'MOEX', Calendar, ql.Russia),
    ('Russia', 'Settlement', Calendar, ql.Russia),
    ('Russia', None, Calendar, ql.Russia),
    ('SaudiArabia', 'Tadawul', Calendar, ql.SaudiArabia),
    ('SaudiArabia', 'Settlement', Calendar, ql.SaudiArabia),
    ('SaudiArabia', None, Calendar, ql.SaudiArabia),
    ('Singapore', 'SGX', Calendar, ql.Singapore),
    ('Singapore', 'Settlement', Calendar, ql.Singapore),
    ('Singapore', None, Calendar, ql.Singapore),
    ('Slovakia', 'BSSE', Calendar, ql.Slovakia),
    ('Slovakia', 'Settlement', Calendar, ql.Slovakia),
    ('Slovakia', None, Calendar, ql.Slovakia),
    ('SouthAfrica', 'Settlement', Calendar, ql.SouthAfrica),
    ('SouthAfrica', None, Calendar, ql.SouthAfrica),
    ('SouthKorea', 'KRX', Calendar, ql.SouthKorea),
    ('SouthKorea', 'Settlement', Calendar, ql.SouthKorea),
    ('SouthKorea', None, Calendar, ql.SouthKorea),
    ('Sweden', 'Settlement', Calendar, ql.Sweden),
    ('Sweden', None, Calendar, ql.Sweden),
    ('Switzerland', 'Settlement', Calendar, ql.Switzerland),
    ('Switzerland', None, Calendar, ql.Switzerland),
    ('Taiwan', 'TSEC', Calendar, ql.Taiwan),
    ('Taiwan', 'Settlement', Calendar, ql.Taiwan),
    ('Taiwan', None, Calendar, ql.Taiwan),
    ('TARGET', 'Settlement', Calendar, ql.TARGET),
    ('TARGET', None, Calendar, ql.TARGET),
    ('Thailand', 'Settlement', Calendar, ql.Thailand),
    ('Thailand', None, Calendar, ql.Thailand),
    ('Turkey', 'Settlement', Calendar, ql.Turkey),
    ('Turkey', None, Calendar, ql.Turkey),
    ('Ukraine', 'USE', Calendar, ql.Ukraine),
    ('Ukraine', 'Settlement', Calendar, ql.Ukraine),
    ('Ukraine', None, Calendar, ql.Ukraine),
    ('UnitedKingdom', 'Exchange', Calendar, ql.UnitedKingdom),
    ('UnitedKingdom', 'Metals', Calendar, ql.UnitedKingdom),
    ('UnitedKingdom', 'Settlement', Calendar, ql.UnitedKingdom),
    ('UnitedKingdom', None, Calendar, ql.UnitedKingdom),
    ('UnitedStates', 'FederalReserve', Calendar, ql.UnitedStates),
    ('UnitedStates', 'GovernmentBond', Calendar, ql.UnitedStates),
    ('UnitedStates', 'LiborImpact', Calendar, ql.UnitedStates),
    ('UnitedStates', 'NERC', Calendar, ql.UnitedStates),
    ('UnitedStates', 'NYSE', Calendar, ql.UnitedStates),
    ('UnitedStates', 'Settlement', Calendar, ql.UnitedStates),
    ('UnitedStates', None, Calendar, ql.UnitedStates),
    ('WeekendsOnly', 'Settlement', Calendar, ql.WeekendsOnly),
    ('WeekendsOnly', None, Calendar, ql.WeekendsOnly),
    (None, None, Calendar, ql.TARGET)  # default behavior
])
def test_static_data_calendar(region: Optional[str], market: Optional[str],
                              expected_class_type: type,
                              expected_ql_class_type: type) -> None:
    # Arrange
    my_calendar = Calendar(region=region, market=market)

    # Act
    my_static_data = StaticData(calendar=my_calendar)

    # Assert
    assert isinstance(my_static_data.calendar, expected_class_type)
    assert isinstance(my_static_data.get_ql_calendar(), expected_ql_class_type)

    # default settings
    assert isinstance(my_static_data.day_counter, my_default_day_counter_type())
    assert isinstance(my_static_data.business_day_convention, my_default_business_day_convention_type())


# endregion Calendar

# region BusinessDayConvention
@pytest.mark.parametrize('business_day_convention, expected_class_type, expected_ql_value', [
    ('Following', BusinessDayConvention, ql.Following),
    ('ModifiedFollowing', BusinessDayConvention, ql.ModifiedFollowing),
    ('Preceding', BusinessDayConvention, ql.Preceding),
    ('ModifiedPreceding', BusinessDayConvention, ql.ModifiedPreceding),
    ('Unadjusted', BusinessDayConvention, ql.Unadjusted),
    (None, BusinessDayConvention, ql.ModifiedFollowing)  # default behavior
])
def test_static_data_business_day_convention(business_day_convention: Optional[str],
                                             expected_class_type: type,
                                             expected_ql_value: int):
    # Act
    my_static_data = StaticData(business_day_convention=business_day_convention)

    # Assert
    assert isinstance(my_static_data.business_day_convention, expected_class_type)
    assert my_static_data.get_default_ql_business_day_convention() == expected_ql_value

    # default settings
    assert isinstance(my_static_data.day_counter, my_default_day_counter_type())
    assert isinstance(my_static_data.calendar, my_default_calendar_type())


@pytest.mark.parametrize('business_day_convention', [
    'Followin'  # invalid business day convention
])
def test_static_data_invalid_business_day_convention(business_day_convention: Optional[str]):
    with pytest.raises(Exception) as e:
        _ = StaticData(business_day_convention=business_day_convention)

    # Assert
    assert str(e.value) == f'Invalid business day convention \'{business_day_convention}\''


# endregion

# region from_json
def test_static_data_from_json():
    # Arrange
    my_json = {
        'day_counter': 'Actual360',
        'calendar': {
            'region': 'UnitedStates',
            'market': 'NYSE',
        },
        'business_day_convention': 'ModifiedFollowing'
    }

    # Act
    my_static_data = StaticData.from_json(my_json)

    # Assert
    assert isinstance(my_static_data.business_day_convention, BusinessDayConvention)
    assert isinstance(my_static_data.day_counter, DayCounter)
    assert isinstance(my_static_data.calendar, Calendar)


# endregion

# region to_json

@pytest.mark.parametrize('day_counter, business_day_convention, region, market', [
    ('Actual360', 'Following', 'UnitedStates', 'Settlement'),
    (None, None, None, None)
])
def test_static_data_to_json(day_counter: Union[str, None], business_day_convention: Union[str, None],
                             region: Union[str, None], market: Union[str, None]):
    # Arrange
    my_calendar = Calendar(region=region, market=market)
    my_static_data = StaticData(day_counter=day_counter, business_day_convention=business_day_convention,
                                calendar=my_calendar)

    # Act
    my_json = my_static_data.to_json()

    # Assert
    assert isinstance(my_json, dict)
    assert 'day_counter' in my_json
    assert 'calendar' in my_json
    assert 'business_day_convention' in my_json
    if day_counter is None:
        assert my_json['day_counter'] == StaticData.get_default_day_counter().name
    else:
        assert my_json['day_counter'] == day_counter
    if business_day_convention is None:
        assert my_json['business_day_convention'] == StaticData.get_default_business_day_convention().name
    else:
        assert my_json['business_day_convention'] == business_day_convention
    assert my_json['calendar']['region'] == region
    assert my_json['calendar']['market'] == market

# endregion



from datetime import datetime
from typing import Union

import QuantLib as ql
from marshmallow import Schema, fields, post_load

from exotx.enums.enums import RandomNumberGenerator
from exotx.helpers.dates import convert_maturity_to_ql_date
from exotx.instruments.basket_type import BasketType, convert_basket_type, BasketTypeField
from exotx.instruments.instrument import Instrument
from exotx.instruments.option_type import convert_option_type_to_ql, OptionType, OptionTypeField
from exotx.utils.pricing_configuration import PricingConfiguration


class BasketOption(Instrument):
    def __init__(self,
                 strike: float,
                 maturity: Union[str, datetime],
                 option_type: Union[str, OptionType],
                 basket_type: Union[str, BasketType]):
        assert strike >= 0, "Invalid strike: cannot be negative"
        self.strike = strike
        self.maturity = convert_maturity_to_ql_date(maturity)
        self.option_type = convert_option_type_to_ql(option_type)
        self.basket_type = convert_basket_type(basket_type)

    def price(self, market_data, static_data, pricing_config: PricingConfiguration, seed: int = 1) -> dict:
        # set the reference date
        reference_date: ql.Date = market_data.get_ql_reference_date()
        ql.Settings.instance().evaluationDate = reference_date

        # create the product
        ql_payoff = ql.PlainVanillaPayoff(self.option_type, self.strike)
        ql_exercise = ql.EuropeanExercise(self.maturity)

        ql_basket_payoff = self._basket_type_to_payoff(ql_payoff)
        ql_option = ql.BasketOption(ql_basket_payoff, ql_exercise)

        # set the pricing engine
        ql_engine = self._get_ql_pricing_engine(market_data, static_data, pricing_config, seed)
        ql_option.setPricingEngine(ql_engine)

        # price
        price = ql_option.NPV()
        if pricing_config.compute_greeks:
            delta = ql_option.delta()
            gamma = ql_option.gamma()
            theta = ql_option.theta()
            return {'price': price, 'delta': delta, 'gamma': gamma, 'theta': theta}
        else:
            return {'price': price}

    def _basket_type_to_payoff(self, payoff: ql.PlainVanillaPayoff):
        """
        Converts a plain vanilla payoff to a basket payoff based on the basket type.

        Args:
            payoff (ql.PlainVanillaPayoff): A plain vanilla payoff object from QuantLib.

        Returns:
            ql.BasketPayoff: A QuantLib BasketPayoff object corresponding to the specified basket type.

        Raises:
            Exception: If the basket type is not one of the supported types (MINBASKET, MAXBASKET, SPREADBASKET, AVERAGEBASKET).
        """
        if self.basket_type == BasketType.MINBASKET:
            return ql.MinBasketPayoff(payoff)
        elif self.basket_type == BasketType.MAXBASKET:
            return ql.MaxBasketPayoff(payoff)
        elif self.basket_type == BasketType.SPREADBASKET:
            return ql.SpreadBasketPayoff(payoff)
        elif self.basket_type == BasketType.AVERAGEBASKET:
            return ql.AverageBasketPayoff(payoff)
        else:
            raise Exception("Invalid basket type")

    @staticmethod
    def _get_ql_pricing_engine(market_data, static_data, pricing_config: PricingConfiguration, seed: int):
        """
        Constructs a QuantLib MCEuropeanBasketEngine for pricing basket options using Monte Carlo simulations.

        This function creates a list of Black-Scholes-Merton processes for each underlying asset, using the spot prices
        and volatilities provided by the MarketData instance. The processes are then combined into a StochasticProcessArray,
        taking into account the correlation matrix from the MarketData instance.

        Finally, a MCEuropeanBasketEngine is constructed using the StochasticProcessArray, along with the specified
        random number generator, time steps per year, required samples, and random seed.

        Args:
            market_data (MarketData): A MarketData instance containing the required market data, including underlying
                                      spot prices, volatilities, and the correlation matrix.
            static_data (StaticData): A StaticData instance containing static information such as calendar and day counter.
            seed (int): A random seed for the Monte Carlo simulations.

        Returns:
            ql.MCEuropeanBasketEngine: A QuantLib MCEuropeanBasketEngine for pricing basket options.

        Note:
            The current implementation only supports MCEuropeanBasketEngine. Additional pricing engines can be added
            based on the basket type and/or pricing configuration.
        """

        # set the reference date
        reference_date = market_data.get_ql_reference_date()

        # set static data
        calendar = static_data.get_ql_calendar()
        day_counter = static_data.get_ql_day_counter()

        processes = [ql.BlackScholesMertonProcess(ql.QuoteHandle(ql.SimpleQuote(x)),
                                                  market_data.get_dividend_curve(day_counter),
                                                  market_data.get_yield_curve(day_counter),
                                                  ql.BlackVolTermStructureHandle(
                                                      ql.BlackConstantVol(reference_date, calendar, y, day_counter)))
                     for x, y in zip(market_data.underlying_spots, market_data.underlying_black_scholes_volatilities)]
        multi_processes = ql.StochasticProcessArray(processes, market_data.get_correlation_matrix())

        # TODO: Consider different pricing engines based on self.basket_type and/or pricing_config
        return ql.MCEuropeanBasketEngine(multi_processes, RandomNumberGenerator.PSEUDORANDOM.value, timeStepsPerYear=1,
                                         requiredSamples=100000, seed=seed)


# region Schema
class BasketOptionSchema(Schema):
    strike = fields.Float()
    maturity = fields.Date(format="%Y-%m-%d")
    option_type = OptionTypeField()
    basket_type = BasketTypeField()

    @post_load
    def make_basket_option(self, data, **kwargs) -> BasketOption:
        return BasketOption(**data)
# endregion



from datetime import datetime
from enum import Enum

import QuantLib as ql

from exotx.data.marketdata import MarketData
from exotx.data.staticdata import StaticData
from exotx.instruments.instrument import Instrument
from exotx.models.blackscholesmodel import BlackScholesModel
from exotx.models.hestonmodel import HestonModel


class OptionType(Enum):
    CALL = 'call'
    PUT = 'put'


class BarrierType(Enum):
    DOWNANDIN = 'downandin'
    DOWNANDOUT = 'downandout'
    UPANDIN = 'upandin'
    UPANDOUT = 'upandout'


class ExerciseType(Enum):
    EUROPEAN = 'european'
    AMERICAN = 'american'


class BarrierOptionEngine(Enum):
    ANALYTICBARRIERENGINE = 'analytic'
    FDBLACKSCHOLESBARRIERENGINE = 'fd-bs-barrier'
    FDBLACKSCHOLESREBATEENGINE = 'fd-bs-rebate'
    FDHESTONBARRIERENGINE = 'fd-heston-barrier'


class BarrierOption(Instrument):
    """Class for modeling a barrier option."""

    def __init__(self,
                 barrier_type: str,
                 barrier: float,
                 strike: float,
                 maturity: str,
                 exercise: str = 'european',
                 option_type: str = 'call',
                 rebate: float = 0):
        self.barrier_type = BarrierType[barrier_type.upper()]
        self.barrier = barrier
        self.strike = strike
        self.maturity = ql.Date().from_date(datetime.strptime(maturity, "%Y-%m-%d"))
        self.exercise = ExerciseType[exercise.upper()]
        self.option_type = OptionType[option_type.upper()]
        self.rebate = rebate
        self.reference_date = ql.Date().todaysDate()
        self.model = None

    def price(self, market_data: MarketData, static_data: StaticData, model: str):
        self.reference_date: ql.Date = market_data.get_ql_reference_date()
        ql.Settings.instance().evaluationDate = self.reference_date

        # create product
        ql_barrier_type = self._get_ql_barrier_type()
        ql_payoff = self._get_ql_payoff()
        ql_exercise = self._get_ql_exercise()
        ql_option = ql.BarrierOption(ql_barrier_type, self.barrier, self.rebate, ql_payoff, ql_exercise)

        # set pricing engine
        ql_pricing_engine = self._get_ql_pricing_engine(market_data, static_data, model)
        ql_option.setPricingEngine(ql_pricing_engine)

        return ql_option.NPV()

    def _get_ql_barrier_type(self) -> ql.Barrier:
        if self.barrier_type == BarrierType.UPANDIN:
            return ql.Barrier.UpIn
        elif self.barrier_type == BarrierType.UPANDOUT:
            return ql.Barrier.UpOut
        elif self.barrier_type == BarrierType.DOWNANDIN:
            return ql.Barrier.DownIn
        else:
            return ql.Barrier.DownOut

    def _get_ql_pricing_engine(self, market_data: MarketData, static_data: StaticData, model: str):
        model = model.lower()
        assert model in [engine.value for engine in BarrierOptionEngine]
        engine = BarrierOptionEngine(model)

        if engine == BarrierOptionEngine.ANALYTICBARRIERENGINE:
            bs_model = BlackScholesModel(market_data, static_data)
            process = bs_model.setup()
            return ql.AnalyticBarrierEngine(process)
        elif engine == BarrierOptionEngine.FDBLACKSCHOLESBARRIERENGINE:
            bs_model = BlackScholesModel(market_data, static_data)
            process = bs_model.setup()
            return ql.FdBlackScholesBarrierEngine(process)
        elif engine == BarrierOptionEngine.FDBLACKSCHOLESREBATEENGINE:
            bs_model = BlackScholesModel(market_data, static_data)
            process = bs_model.setup()
            return ql.FdBlackScholesRebateEngine(process)
        elif engine == BarrierOptionEngine.FDHESTONBARRIERENGINE:
            heston_model = HestonModel(market_data, static_data)
            _, model = heston_model.calibrate()
            return ql.FdHestonBarrierEngine(model)
        else:
            raise NotImplementedError

    def _get_ql_option_type(self):
        if self.option_type == OptionType.CALL:
            return ql.Option.Call
        else:
            return ql.Option.Put

    def _get_ql_payoff(self) -> ql.Payoff:
        ql_option_type = self._get_ql_option_type()
        # TODO: Consider different payoffs?
        return ql.PlainVanillaPayoff(ql_option_type, self.strike)

    def _get_ql_exercise(self) -> ql.Exercise:
        if self.exercise == ExerciseType.EUROPEAN:
            return ql.EuropeanExercise(self.maturity)
        else:
            return ql.AmericanExercise(self.reference_date, self.maturity)
